# PhenotypeQuery.py
# MIT LICENSE 2016
# Shaun Harker

class PhenotypeQuery:
  def __init__ (self, database, mgi):
    self.database = database
    c = database.conn.cursor()
    self.set_of_matches = set([ row[0] for row in c.execute('select ParameterIndex from Signatures where MorseGraphIndex=' + str(mgi) + ';')])

  def matches(self):
    """
    Return entire set of matches
    """
    return self.set_of_matches

  def __call__ (self, pi ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return pi in self.set_of_matches




# EssentialParameterNeighbors.py
# Marcio Gameiro
# 2020-09-18
# MIT LICENSE

from DSGRN import *

def essential_network_spec(parameter_graph):
    """Returns the essential network specification"""
    # Get network specification
    net_spec = parameter_graph.network().specification()
    # Get list of node specifications
    nodes_spec = [s.strip() for s in net_spec.strip().split('\n') if s]
    # Number of specs should equal number of nodes
    assert len(nodes_spec) == parameter_graph.dimension()
    # Check if a node spec correspond to an essential node
    essential_node = lambda spec: spec.strip().endswith('E') and spec.count(':') == 2
    # Make all nodes essential
    ess_nodes_spec = [spec if essential_node(spec) else spec + ' : E' for spec in nodes_spec]
    # Get the essential network spec
    ess_net_spec = '\n'.join(ess_nodes_spec)
    return ess_net_spec

def essential_parameters(parameter_graph):
    """Returns list of essential parameters in the parameter graph"""
    # Get the essential network spec
    ess_net_spec = essential_network_spec(parameter_graph)
    # Construct essential network and its parameter graph
    ess_network = Network(ess_net_spec)
    ess_parameter_graph = ParameterGraph(ess_network)
    # Get list of indices of essential parameters embedded
    # in the parameter graph of the original network
    ess_parameters = [] # Essential parameter indices
    for ess_pindex in range(ess_parameter_graph.size()):
        # Get the essential parameter
        ess_par = ess_parameter_graph.parameter(ess_pindex)
        # Get its index in the original parameter graph
        full_pindex = parameter_graph.index(ess_par)
        # Add the index to the list of essential parameters
        ess_parameters.append(full_pindex)
    return ess_parameters

def essential_parameter_neighbors(parameter_graph, level=1):
    """Returns list of essential parameters and its level edge neighbors"""
    # Get list of essential parameters in the parameter graph
    ess_parameters = essential_parameters(parameter_graph)
    # Get list of neighbors of essential parameters
    prev_level_neighbors = ess_parameters # Previous level neighbors
    next_level_neighbors = set() # Next level neighbors
    ess_par_neighbors = set() # Neighbors of essential parameters
    for lev in range(level):
        # Clear next level neighbors
        next_level_neighbors.clear()
        for ess_pindex in prev_level_neighbors:
            # Get neighbors of this essential parameter
            for p_index in parameter_graph.adjacencies(ess_pindex):
                next_level_neighbors.add(p_index)
        # Update previous level neighbors list
        prev_level_neighbors = list(next_level_neighbors)
        # Get number of neighbors before update
        num_neighbors_before = len(ess_par_neighbors)
        # Update essential parameter neighbors set
        ess_par_neighbors.update(prev_level_neighbors)
        # Check if any new neighbor was added
        if len(ess_par_neighbors) == num_neighbors_before:
            break # No more neighbors to add
    # Remove neighbors that are essential parameters
    ess_par_neighbors.difference_update(ess_parameters)
    # Return list of essential parameters and its neighbors
    return ess_parameters, list(ess_par_neighbors)



### SubdomainGraph.py
### Shaun Harker
### 2018-03-25
### MIT LICENSE
#
# Marcio Gameiro
# 2021-03-26

from pychomp2 import *
import datetime
from itertools import product
from math import log2
from DSGRN._dsgrn import *

class SubdomainGraph:

    def complex(self):
        return self.cc

    def diagram(self):
        return self.digraph.adjacencies

    def domains(self):
        def domgen():
            sizes = self.network.domains()
            dom = Domain(sizes)
            while dom.isValid():
                yield dom
                dom.preincrement()
        return domgen()

    def walls(self):
        return ( wall for wall in self.cc(self.D-1) if not self.cc.rightfringe(wall) )

    def box(self,subdom):
        return self.cc.cell_index(subdom, 2**self.D - 1)

    def subdomains(self):
        return ( box for box in self.cc(self.D) if not self.cc.rightfringe(box) )   

    def centerbox(self,dom):
        return self.box([3*i + 1 for i in dom])

    def wallnormal(self,wall):
        shape = self.cc.cell_shape(wall)
        return int(log2(2**self.D - 1 - shape))

    def posterior(self,wall):
        dim = self.wallnormal(wall)
        return self.cc.left(self.cc.left(wall,dim),dim)

    def anterior(self,wall):
        dim = self.wallnormal(wall)
        return self.cc.right(self.cc.right(wall,dim),dim)

    ## labelling

    def leftlabel(self):
        return 1

    def rightlabel(self):
        return 2   

    def isLeftBoundingWall(self, wall):
        normdim = self.wallnormal(wall)
        return self.cc.leftfringe(wall)
    
    def isRightBoundingWall(self, wall):
        normdim = self.wallnormal(wall)
        return self.cc.rightfringe(self.cc.right(wall, normdim))

    def __init__(self, p, logging = None):
        self.p = p
        self.network = self.p.network()
        self.D = self.network.size()
        self.cc = CubicalComplex([ 3*x + 1 for x in self.network.domains()])
        self.digraph = DirectedAcyclicGraph()
        if logging:
            print(datetime.datetime.now().time(), flush=True)
            print("Complex size = " + str(self.cc.size()), flush=True)
            print("Number of top cells = " + str(self.cc.size(self.D)), flush=True)
        for box in self.cc(self.D):
            self.digraph.add_vertex(box, str(self.cc.coordinates(box)))
        if logging:
            print(datetime.datetime.now().time(), flush=True)
            print("Graph vertices constructed", flush=True)

        # construct wall labels

        # Step 1. Center labels
        label = {}
        for dom in self.domains():
            box = self.centerbox(dom)
            for dim in range(0,self.D):
                x = self.p.absorbing(dom, dim, 1)  # 1 -- right
                rightwall = self.cc.right(box, dim)
                label[rightwall] = self.rightlabel() if x else self.leftlabel()
                y = self.p.absorbing(dom, dim, -1)  # -1 -- left
                leftwall = self.cc.left(box, dim)
                label[leftwall] = self.leftlabel() if y else self.rightlabel()

        if logging:
            print(datetime.datetime.now().time(), flush=True)
            print("Computed center labels", flush=True)

        # Step 2. Center Transverse labels
        for dom in self.domains():
            box = self.centerbox(dom)
            for dim in range(0,self.D):
                for wall in self.posterior(self.cc.left(box, dim)), self.anterior(self.cc.right(box, dim)):
                    if self.isLeftBoundingWall(wall):
                        label[wall] = self.rightlabel()
                        continue
                    if self.isRightBoundingWall(wall): 
                        label[wall] = self.leftlabel()
                        continue
                    posteriorlabel = label.get(self.posterior(wall),0)
                    anteriorlabel = label.get(self.anterior(wall),0)
                    if posteriorlabel == anteriorlabel:
                        for adjwall in self.cc.parallelneighbors(wall):
                            label[adjwall] = posteriorlabel

        if logging:
            print(datetime.datetime.now().time(), flush=True)
            print("Computed transverse labels", flush=True)

        # Step 4. Propagate labels
        newlabel = dict(label)
        for wall in label:
            for adjwall in self.cc.parallelneighbors(wall):
                if self.cc.rightfringe(adjwall): continue
                if adjwall not in newlabel:
                    newlabel[adjwall] = 0
                newlabel[adjwall] |= label[wall]
        label = newlabel 

        if logging:
            print(datetime.datetime.now().time(), flush=True)
            print("Computed all wall labels", flush=True)

        # Build graph edges
        for wall in self.cc(self.D-1):
            #print(" wall = " + str(wall),flush=True)
            normdim = self.wallnormal(wall)
            leftbox = self.cc.left(wall, normdim)
            rightbox = self.cc.right(wall, normdim)
            #print(str(self.cc.coordinates(leftbox)) + " and " + str(self.cc.coordinates(rightbox)))
            #print(label.get(wall,0))
            if label.get(wall,0) in [0,3]:
                #print(str(self.cc.coordinates(leftbox)) + " <-> " + str(self.cc.coordinates(rightbox)))
                self.digraph.add_edge(leftbox, rightbox)
                self.digraph.add_edge(rightbox, leftbox)
            elif label[wall] == self.rightlabel():
                #print(str(self.cc.coordinates(leftbox)) + " -> " + str(self.cc.coordinates(rightbox)))
                self.digraph.add_edge(leftbox, rightbox)
            elif label[wall] == self.leftlabel():
                #print(str(self.cc.coordinates(leftbox)) + " <- " + str(self.cc.coordinates(rightbox)))
                self.digraph.add_edge(rightbox, leftbox)
        if logging:
            print(datetime.datetime.now().time(), flush=True)
            print("Constructed SubdomainGraph", flush=True)



# AlignmentGraph.py
# MIT LICENSE 2016
# Shaun Harker

from DSGRN.Query.Graph import *

def AlignmentGraph ( graph1, graph2, matching_relation ):
  """
  Creates alignment graph of graph1 and graph2 (i.e. induced subgraph of tensor product where "matching_label" attribute matches)
  The graph structures must have: 
    an attribute "vertices" which is the set of vertices,
    a method "adjacency" which gives adjacency lists,
    an attribute "matching_label" so "matching_label(v)" gives a label for each vertex
  """
  vertices = set([ (u, v) for u in graph1.vertices for v in graph2.vertices if matching_relation(graph1.matching_label(u),graph2.matching_label(v)) ])
  edges = [ ((x,y),(u,v)) for (x,y) in vertices for u in graph1.adjacencies(x) for v in graph2.adjacencies(y) if (u,v) in vertices ]
  return Graph(vertices, edges)



# MonostableFixedPointQuery.py
# MIT LICENSE 2016
# Shaun Harker

from DSGRN.Query.SingleFixedPointQuery import *
from DSGRN.Query.MonostableQuery import *
from DSGRN.Query.Logging import LogToSTDOUT

class MonostableFixedPointQuery:
  """
  MonostableFixedPointQuery
    Enables queries to determine if a Morse Graph Index 
    corresponds to a Morse Graph for which there is
    (a) a single minimal Morse node
    (b) the Morse node is annotated as an FP within specified bounds
  """
  def __init__(self, database, bounds):
    LogToSTDOUT("MonostableFixedPointQuery :: initializing")
    self.set_of_matches = SingleFixedPointQuery(database,bounds).matches() & MonostableQuery(database).matches()
    LogToSTDOUT("MonostableFixedPointQuery :: constructed")


  def matches(self):
    """
    Return entire set of matches
    """
    return self.set_of_matches

  def __call__(self, mgi):
    """ 
    Test if a single mgi is in the set of matches
    """
    return mgi in self.set_of_matches


# ParameterFromSample.py
# Marcio Gameiro and Lun Zhang
# MIT LICENSE
# 2021-10-14

import DSGRN
import numpy as np
import json

def partial2hex(partial_order, n_inputs, n_outputs):
    """Computes the hex code corresponding to a given partial order
    The values in between the thresholds do not need to be sorted
    """
    par_bit_matrix = np.zeros([n_outputs, 2**n_inputs], dtype=int) # matrix of bit codes
    for k in range(n_outputs): # for each threshold
        indx = partial_order.index(-(k+1)) # index of kth threshold
        indx_left = [n for n in partial_order[:indx] if n >= 0] # indices >= 0 before threshold
        bit_array = np.ones(2**n_inputs, dtype=int) # array of bit codes for kth threshold
        bit_array[indx_left] = 0 # set entries corresponding to indx_left to zero
        par_bit_matrix[k] = list(reversed(bit_array)) # reverse and set as kth row
    par_bit_code = ''.join(map(str, par_bit_matrix.flatten('F'))) # flatten in column-major order (Fortran-style)
    n_nibbles = ((len(par_bit_code) - 1) // 4) + 1 # number of nibbles (ceil division by 4)
    fmt = '0' + str(n_nibbles) + 'X' # hex code format
    par_hex_code = format(int(par_bit_code, 2), fmt) # set parameter hex code
    return par_hex_code

def indexMask(logic):
    """Generate map i -> bit mask of i for value logic"""
    lgt = sum(logic)
    pattern = '{0:0' + str(lgt) + 'b}'
    n_orders = np.prod([2**s for s in logic])
    ret = []
    for i in range(n_orders):
        mask = pattern.format(i)
        vMask = [int(i) for i in mask]
        currMask = []
        currPos = 0
        for i in range(len(logic)):
            temp = vMask[currPos:currPos+logic[i]]
            currMask.append(temp)
            currPos += logic[i]
        ret.append(currMask)
    return ret

def evalMask(sample, masks):
    """Given masks evaluate the sample sequence of switching polynomials"""
    ret = []
    for i in range(len(masks)):
        currVal = 1
        currPos = 0
        for j in range(len(masks[i])):
            tempSum = 0
            for k in range(len(masks[i][j])):
                if masks[i][j][k] == 0:
                    tempSum += sample[currPos][0]
                else:
                    tempSum += sample[currPos][1]
                currPos += 1
            currVal *= tempSum
        ret.append(currVal)
    return ret

def binSort(x):
    prev = 0
    for i in range(len(x)):
        if x[i] < 0:
            x[prev:i] = sorted(x[prev:i])
            prev = i + 1
    lgt = len(x)
    x[prev:lgt] = sorted(x[prev:lgt])
    return x

def nodeRegion(logic, samples, thetas):
    masks = indexMask(logic)
    tempVals = evalMask(samples,masks)
    if len(np.unique(tempVals)) != len(tempVals):
        return None
    rthetas = thetas.copy()[::-1]
    tempOrd = np.argsort(tempVals+rthetas)
    for i in range(len(tempOrd)):
        if tempOrd[i] >= len(tempVals):
            tempOrd[i] -= len(tempOrd)
    tempOrd = str(list(binSort(tempOrd)))
    return eval(tempOrd)

def par_index_from_sample_old(parameter_graph, L, U, T):
    network = parameter_graph.network()
    D = network.size()
    # The logic of each node, logics[d], is a list of the factor lengths
    logics = [[len(logic) for logic in network.logic(d)] for d in range(D)]
    logic_parameters = []
    order_parameters = []
    for d in range(D):
        # Input edges from source s to target d
        # Output edges from source d to target s
        inputs = network.inputs(d)
        inputs.sort(reverse=True)
        n_inputs = len(inputs)
        outputs = network.outputs(d)
        outputs.sort(reverse=True)
        n_outputs = len(outputs)
        L_U_values = [[L[s, d], U[s, d]] for s in inputs]
        T_values = [T[d, s] for s in outputs]
        node_region = nodeRegion(logics[d], L_U_values, T_values)
        # Return -1 for invalid input
        if node_region is None:
            return -1
        # Get the thresholds from node_region
        thres = [p for p in node_region if p < 0]
        # Get thresholds permutation (order parameter)
        order = [p + n_outputs for p in thres]
        # Get order parameter
        order_parameters.append(DSGRN.OrderParameter(order))
        # Add sorted thresholds (in increasing order) to partial order
        part_order = [thres.index(p) - n_outputs if p < 0 else p for p in node_region]
        # Get hex code from partial order
        hex_code = partial2hex(part_order, n_inputs, n_outputs)
        # Get logic parameter
        logic_parameters.append(DSGRN.LogicParameter(n_inputs, n_outputs, hex_code))
    # Get DSGRN parameter
    parameter = DSGRN.Parameter(logic_parameters, order_parameters, network)
    # Get parameter index from parameter graph
    par_index = parameter_graph.index(parameter)
    # Return parameter index if valid, else return -1
    if par_index < parameter_graph.size():
        return par_index
    return -1

def index_from_partial_orders(parameter_graph, partial_orders):
    network = parameter_graph.network()
    D = network.size()
    logic_parameters = []
    order_parameters = []
    for d in range(D):
        n_inputs = len(network.inputs(d))
        n_outputs = len(network.outputs(d))
        # Get partial order for this node
        partial_order = partial_orders[d]
        # Transform into list if is a string
        if isinstance(partial_order, str):
            # Remove white space and parentheses and split string into a list
            partial_order = partial_order.replace(' ', '')[1:-1].split(',')
        # Extract numerical values (partial order can be numeric or string)
        get_digit = lambda p : int(p[1:]) if p[0] == 'p' else int(p[1:]) - n_outputs
        # Partial order for this node (with threshold permutations)
        part_order = [p if isinstance(p, int) else get_digit(p) for p in partial_order]
        # Get the thresholds from partial_order
        thres = [p for p in part_order if p < 0]
        # Get thresholds permutation (order parameter)
        order = [p + n_outputs for p in thres]
        # Get order parameter
        order_parameters.append(DSGRN.OrderParameter(order))
        # Add sorted thresholds (in increasing order) to partial order
        partial_order = [thres.index(p) - n_outputs if p < 0 else p for p in part_order]
        # Get hex code from partial order
        hex_code = partial2hex(partial_order, n_inputs, n_outputs)
        # Get logic parameter
        logic_parameters.append(DSGRN.LogicParameter(n_inputs, n_outputs, hex_code))
    # Get DSGRN parameter
    parameter = DSGRN.Parameter(logic_parameters, order_parameters, network)
    # Get parameter index from parameter graph
    par_index = parameter_graph.index(parameter)
    # Return parameter index if valid, else return -1
    if par_index < parameter_graph.size():
        return par_index
    return -1

def par_index_from_sample(parameter_graph, L, U, T):
    network = parameter_graph.network()
    D = network.size()
    # The logic of each node, logics[d], is a list of the factor lengths
    logics = [[len(logic) for logic in network.logic(d)] for d in range(D)]
    partial_orders = []
    for d in range(D):
        # Input edges from source s to target d
        # Output edges from source d to target s
        inputs = network.inputs(d)
        inputs.sort(reverse=True)
        n_inputs = len(inputs)
        outputs = network.outputs(d)
        outputs.sort(reverse=True)
        n_outputs = len(outputs)
        L_U_values = [[L[s, d], U[s, d]] for s in inputs]
        T_values = [T[d, s] for s in outputs]
        node_region = nodeRegion(logics[d], L_U_values, T_values)
        # Return -1 for invalid input
        if node_region is None:
            return -1
        partial_orders.append(node_region)
    par_index = index_from_partial_orders(parameter_graph, partial_orders)
    return par_index



# InducibilityQuery.py
# MIT LICENSE 2016
# Shaun Harker and Bree Cummins

from DSGRN.Query.SingleFixedPointQuery import *
from DSGRN.Query.DoubleFixedPointQuery import *
from DSGRN.Query.MonostableFixedPointQuery import *
from DSGRN.Query.SingleGeneQuery import *

class InducibilityQuery:
  """
  InducibilityQuery. 
    The query depends on a choice of gene, and two sets of domain bounds named bounds1 and bounds2.
    Given: a reduced parameter index 
    Return: a triple of boolean values (a,b,c) such that:
      a is true if and only if (0,reduced_parameter_index) is monostable and has an FP in bounds1
      b is true if and only if (max_gpi,reduced_parameter_index) is monostable and has an FP in bounds2
      c is true if and only if (gpi,reduced_parameter_index) has an FP in bounds1 and an FP in bounds2
                               for some 0 < gpi < max_gpi
  """
  def __init__(self, database, gene, bounds1, bounds2):
    self.QueryFP1 = MonostableFixedPointQuery(database, bounds1)
    self.QueryFP2 = MonostableFixedPointQuery(database, bounds2)
    self.QueryDoubleFP = DoubleFixedPointQuery(database,bounds1,bounds2)
    self.GeneQuery = SingleGeneQuery(database, gene)
    self.max_gpi = self.GeneQuery.number_of_gene_parameters() - 1

  def __call__(self, reduced_parameter_index):
    """
    Given a reduced parameter index, return a triple of boolean values (a,b,c)
    """
    query_graph = self.GeneQuery(reduced_parameter_index)
    a = self.QueryFP1(query_graph.mgi(0))
    b = self.QueryFP2(query_graph.mgi(self.max_gpi))
    c = any ( self.QueryDoubleFP(query_graph.mgi(i)) for i in range(1,self.max_gpi) )
    return (a, b, c)


# SingleGeneQuery.py
# MIT LICENSE 2016
# Shaun Harker
#
# Marcio Gameiro
# 2021-03-26

from DSGRN.Query.Graph import *
from DSGRN.Query.Hexcodes import *
from DSGRN.Query.Logging import LogToSTDOUT

class SingleGeneQuery:
  def __init__(self, database, gene):
    LogToSTDOUT("SingleGeneQuery(" + str(database.dbname) + ", " + str(gene) + ")")
    self.database = database
    self.gene = gene
    # "gene_index" gives the integer index used in the representation
    self.gene_index = self.database.network.index(self.gene)

    # num_gene_param is the size of the factor graph associated with "gene"
    self.num_gene_param = database.indexing_place_bases[self.gene_index]
    # num_reduced_param is the product of the sizes of all remaining factor graphs, and reorderings of all genes (including "gene")
    self.num_reduced_param = database.parametergraph.size() // self.num_gene_param

    # Create factor graph
    # TODO: The following algorithm for creating a factor graph takes O(NM^2) time, where N is length of hexcodes and M is size of factorgraph.
    #       An O(NM) algorithm is possible (based on looping through codewords and trying out each promotion, using a hash table)
    #       Implement this.
    self.hexcodes = self.database.parametergraph.factorgraph(self.gene_index)
    vertices = set(range(0,len(self.hexcodes)))
    edges = [ (gpi1, gpi2) for gpi1 in vertices for gpi2 in vertices if isAdjacentHexcode(self.hexcodes[gpi1], self.hexcodes[gpi2]) ]
    self.graph = Graph(vertices,edges)
    
    LogToSTDOUT("SingleGeneQuery: FactorGraph generated")

    if not hasattr(database, 'SingleGeneQuery'):
      LogToSTDOUT("SingleGeneQuery: SingleGeneQuery attribute missing from python database object.")
      database.SingleGeneQuery = set()
      # For single-gene manipulations, we convert from parameter index to (reduced parameter index, gene parameter index)
      # To recover the parameter index from the reduced parameter index, we insert a "digit":
      database.full_parameter_index = lambda rpi, gpi, gene_index : rpi % database.indexing_place_values[gene_index] + gpi * database.indexing_place_values[gene_index] + (rpi // database.indexing_place_values[gene_index]) * database.indexing_place_values[gene_index+1]
      database.reduced_parameter_index = lambda pi, gene_index : (pi % database.indexing_place_values[gene_index] + (pi // database.indexing_place_values[gene_index+1]) * database.indexing_place_values[gene_index], (pi // database.indexing_place_values[gene_index]) % database.indexing_place_bases[gene_index] )
    
    LogToSTDOUT("SingleGeneQuery: SingleGeneQuery attributes created.")

    if gene not in database.SingleGeneQuery:
      LogToSTDOUT("SingleGeneQuery: database structure unaware of gene " + str(gene) )
      # Sanitize "gene":
      if gene not in database.names:
        raise NameError(gene + " is not the name of a node in the network")

      LogToSTDOUT("SingleGeneQuery: sanitized " + str(gene) )
      # Now we scan the database and create a new table we can do single gene poset queries with
      c = database.conn.cursor()
      # c.execute("drop table if exists " + gene) # gene is sanitized
      # Check if table exists
      LogToSTDOUT("SingleGeneQuery: cursor constructed " )

      c.execute("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='" + gene + "';")
      LogToSTDOUT("SingleGeneQuery: checked for table" )

      table_exists = c.fetchone()[0]
      if not table_exists:
        LogToSTDOUT("SingleGeneQuery: table did not exist" )
        c.execute("create table " + gene + " (ReducedParameterIndex INTEGER, GeneParameterIndex INTEGER, ParameterIndex INTEGER, MorseGraphIndex INTEGER)")
        LogToSTDOUT("SingleGeneQuery: created table" )

        # # old method -- loads whole table into memory
        # values = [ self.reduced_parameter_index(row[0], self.gene_index) + row for row in c.execute("select * from Signatures")]
        # c.executemany('INSERT INTO ' + gene + ' VALUES (?,?,?,?)', values )

        # new method -- a generator that uses fetchmany to keep memory usage down
        # default value arraysize=10000 was chosen based on limited testing. No improvement seen at 50000 or 100000 on a 13M parameter db. Worse performance at arraysize=1000.
        def valuegenerator (cursor,arraysize=10000):
          while True:
              rows = cursor.fetchmany(arraysize)
              if not rows:
                  break
              yield [database.reduced_parameter_index(row[0], self.gene_index) + row for row in rows]

        c2 = database.conn.cursor() # need a second cursor to keep the place in the iteration as updates to the table occur
        c2.execute("select * from Signatures")
        LogToSTDOUT("SingleGeneQuery: executed select statement on signatures" )
        for values in valuegenerator(c2):
          c.executemany('INSERT INTO ' + gene + ' VALUES (?,?,?,?)', values )
        LogToSTDOUT("SingleGeneQuery: inserted values" )
        c2.close()
        LogToSTDOUT("SingleGeneQuery: closed cursor" )
        c.execute('create index ' + gene + '1 on ' + gene + ' (ReducedParameterIndex, GeneParameterIndex)') # gene is sanitized
        LogToSTDOUT("SingleGeneQuery: created index" )
        database.conn.commit()
        LogToSTDOUT("SingleGeneQuery: committed database transaction" )
      database.SingleGeneQuery.add(gene)
      LogToSTDOUT("SingleGeneQuery: added gene to python database object.")
    LogToSTDOUT("SingleGeneQuery: constructed" )


  def __call__ (self, reduced_parameter_index):
    """
    The query returns a graph which contains the poset of gene parameter indices 
    corresponding to adjusting the parameter by changing the logic parameter associated 
    with the gene being queried.
    The graph is as follows: 

    * The vertices of the graph are named according to Gene Parameter Index (gpi). 
    * There is a directed edge p -> q iff p < q and the associated logic parameters are adjacent.
    * The graph is labelled with pairs (Parameter index, Morse graph index).

    In addition the following extra structures are provided:

    * `graph.data` is a dictionary from gene parameter index to (hex code, parameter index, morse graph index)
    * `graph.mgi` is a function which accepts a gpi and returns the associated Morse graph index
    * `graph.num_inputs` is the number of network edges which are inputs to the gene associated with the query
    * `graph.num_outputs`is the number of network edges which are outputs to the gene associated with the query
    * `graph.essential` is a boolean-valued function which determines if each vertex corresponds to an essential parameter node
    """
    LogToSTDOUT("SingleGeneQuery(" + str(reduced_parameter_index) + ")")
    c = self.database.execute("select GeneParameterIndex,ParameterIndex,MorseGraphIndex from " + self.gene + " where ReducedParameterIndex=?", (str(reduced_parameter_index),))
    LogToSTDOUT("SingleGeneQuery: SQL statement executed")
    Q = { row[0] : (self.hexcodes[row[0]],  row[1], row[2]) for row in c }
    LogToSTDOUT("SingleGeneQuery: Q constructed")
    graph = self.graph.clone()
    LogToSTDOUT("SingleGeneQuery: graph constructed")
    graph.data = Q
    graph.mgi = lambda gpi : graph.data[gpi][2]
    graph.label = lambda gpi : str(graph.data[gpi][1]) + ':' + str(graph.data[gpi][2])
    graph.num_inputs = len(self.database.network.inputs(self.gene_index))
    graph.num_outputs = len(self.database.network.outputs(self.gene_index))
    graph.essential = lambda gpi : essential(graph.data[gpi][0],graph.num_inputs,graph.num_outputs)
    LogToSTDOUT("SingleGeneQuery: graph attributes emplaced")
    return graph

  def number_of_gene_parameters(self):
    """
    Return number of gene parameters associated with query object
    This is the size of the logic factor graph associated with the selected gene
    """
    return self.num_gene_param

  def number_of_reduced_parameters(self):
    """
    Return number of reduced parameters associated with query object
    This is the size of the parameter grapph divided by the size of the 
    logic factor graph associated with the selected gene
    """  
    return self.num_reduced_param






# StableFCQuery.py
# MIT LICENSE 2016
# Shaun Harker

from collections import defaultdict

class StableFCQuery:
  def __init__ (self, database):
    self.database = database
    c = database.conn.cursor()
    query_result = [ row for row in c.execute('select ParameterIndex, Vertex from Signatures natural join (select MorseGraphIndex,Vertex from (select MorseGraphIndex,Vertex from MorseGraphAnnotations where Label="FC" except select MorseGraphIndex,Source from MorseGraphEdges));')]
    self.query_data = defaultdict(set)
    for parameter_index, morse_node in query_result:
      self.query_data[parameter_index].add(morse_node)
    #result = [row[0] for row in c.execute('select distinct ParameterIndex from Matches natural join Signatures order by ParameterIndex;')]

  def matches(self):
    """
    Return entire set of matches
    """
    return self.query_data

  def __call__ (self, parameter_index ):
    """ 
    Test if a single parameter index is in the set of matches
    """
    return self.query_data[parameter_index]



# MonostableQuery.py
# MIT LICENSE 2016
# Shaun Harker

from DSGRN.Query.Logging import LogToSTDOUT

class MonostableQuery:
  """
  MonostableQuery
    Query to check if morse graph index corresponds to a Morse graph has a unique minimal element
  """
  def __init__ ( self, database ):
    LogToSTDOUT("MonostableQuery :: initializing")
    self.database = database
    if not hasattr(database, 'MonostableQuery'):
      c = database.conn.cursor()
      LogToSTDOUT("MonostableQuery :: select MorseGraphIndex from (select MorseGraphIndex, count(*) as StableCount from (select MorseGraphIndex,Vertex from MorseGraphVertices except select MorseGraphIndex,Source from MorseGraphEdges) group by MorseGraphIndex) where StableCount=1;")
      sqlexpression = "select MorseGraphIndex from (select MorseGraphIndex, count(*) as StableCount from (select MorseGraphIndex,Vertex from MorseGraphVertices except select MorseGraphIndex,Source from MorseGraphEdges) group by MorseGraphIndex) where StableCount=1;"
      database.MonostableQuery = frozenset([ row[0] for row in c.execute(sqlexpression) ])
    LogToSTDOUT("MonostableQuery :: constructed")

  def matches(self):
    """
    Return entire set of matches
    """
    return self.database.MonostableQuery

  def __call__ ( self, morsegraphindex ):
    """ 
    Test if a single mgi is in the set of matches
    """
    return morsegraphindex in self.database.MonostableQuery



import datetime, sys
def LogToSTDOUT(string=''):
  """
  Print date and time, along with optional string input, to stdout.
  """
  if string: 
    string = ':\n' + str(string)
  else:
    string = ''
  print(str(datetime.datetime.now())+string)
  sys.stdout.flush()



# Signatures

from mpi4py import MPI
from mpi4py.futures import MPICommExecutor
from DSGRN import *
import sqlite3, sys, time
import progressbar # pip install progressbar2

def SaveDatabase(filename, data, pg):
    print("Save Database")
    N = pg.size()
    conn = sqlite3.connect(filename)
    conn.executescript("""
      create table if not exists Signatures (ParameterIndex INTEGER PRIMARY KEY, MorseGraphIndex INTEGER);
      create table if not exists MorseGraphViz (MorseGraphIndex INTEGER PRIMARY KEY, Graphviz TEXT);
      create table if not exists MorseGraphVertices (MorseGraphIndex INTEGER, Vertex INTEGER);
      create table if not exists MorseGraphEdges (MorseGraphIndex INTEGER, Source INTEGER, Target INTEGER);
      create table if not exists MorseGraphAnnotations (MorseGraphIndex INTEGER, Vertex INTEGER, Label TEXT);
      create table if not exists Network ( Name TEXT, Dimension INTEGER, Specification TEXT, Graphviz TEXT);
      """)

    # Postprocessing to give Morse Graphs indices
    morsegraphs = []
    def signatures_table(data):
        bar = progressbar.ProgressBar(max_value=N)
        morsegraphindices = {}
        for (pi, mg) in data:
            bar.update(pi)
            if mg in morsegraphindices: # ideally I'd have a graph isomorphism check
                mgi = morsegraphindices[mg]
            else:
                mgi = len(morsegraphindices)
                morsegraphindices[mg] = mgi
                morsegraphs.append(mg)
            yield (pi,mgi)
        bar.finish()

    def MG(mgi):
        return MorseGraph().parse(morsegraphs[mgi])

    name = filename
    if filename[-3:] == '.db':
        name = filename[:-3]
        
    print("Inserting Network table into Database", flush=True)
    conn.execute("insert into Network ( Name, Dimension, Specification, Graphviz) values (?, ?, ?, ?);", (name, pg.network().size(), pg.network().specification(), pg.network().graphviz()))

    print("Inserting Signatures table into Database", flush=True)
    conn.executemany("insert into Signatures (ParameterIndex, MorseGraphIndex) values (?, ?);", signatures_table(data))

    print("Inserting MorseGraphViz table into Database", flush=True)
    conn.executemany("insert into MorseGraphViz (MorseGraphIndex, Graphviz) values (?, ?);",
      ( (mgi, MG(mgi).graphviz()) for mgi in progressbar.ProgressBar()(range(0, len(morsegraphs))) ))

    print("Inserting MorseGraphVertices table into Database", flush=True)
    conn.executemany("insert into MorseGraphVertices (MorseGraphIndex, Vertex) values (?, ?);",
      ( (mgi, v) for mgi in progressbar.ProgressBar()(range(0, len(morsegraphs))) for v in range(0,MG(mgi).poset().size()) ))

    print("Inserting MorseGraphEdges table into Database", flush=True)
    conn.executemany("insert into MorseGraphEdges (MorseGraphIndex, Source, Target) values (?, ?, ?);",
      ( (mgi, s, t) for mgi in progressbar.ProgressBar()(range(0, len(morsegraphs))) for s in range(0,MG(mgi).poset().size()) for t in MG(mgi).poset().children(s) ))

    print("Inserting MorseGraphAnnotations table into Database", flush=True)
    conn.executemany("insert into MorseGraphAnnotations (MorseGraphIndex, Vertex, Label) values (?, ?, ?);",
      ( (mgi, v, label) for mgi in progressbar.ProgressBar()(range(0, len(morsegraphs))) for v in range(0,MG(mgi).poset().size()) for label in MG(mgi).annotation(v) ))

    print("Indexing Database.", flush=True)
    conn.executescript("""
      create index if not exists Signatures2 on Signatures (MorseGraphIndex, ParameterIndex);
      create index if not exists MorseGraphAnnotations3 on MorseGraphAnnotations (Label, MorseGraphIndex);
      create index if not exists MorseGraphViz2 on MorseGraphViz (Graphviz, MorseGraphIndex);
      create index if not exists MorseGraphVertices1 on MorseGraphVertices (MorseGraphIndex, Vertex);
      create index if not exists MorseGraphVertices2 on MorseGraphVertices (Vertex, MorseGraphIndex);
      create index if not exists MorseGraphEdges1 on MorseGraphEdges (MorseGraphIndex);
      create index if not exists MorseGraphAnnotations1 on MorseGraphAnnotations (MorseGraphIndex);
      """)
    conn.commit()
    conn.close()

# mpi4py seems to require the following to be global for workers to see it

helpstring = """
Signatures
  Usage Example:  
    mpiexec -np 9 --oversubscribe -use-hwthread-cpus \
      Signatures $netspecfile $outputfile
    where $netspecfile is the path to a network specification file
      and $outputfile  is the path to write the SQL database output to.
"""
if len(sys.argv) != 3:
    print(helpstring)
    exit(1)
specfile = sys.argv[1]
outfile = sys.argv[2]
gpg = ParameterGraph(Network(specfile))

def work(pi): 
  return (pi, MorseGraph(DomainGraph(gpg.parameter(pi))).stringify())

def main():
    global gpg
    with MPICommExecutor(MPI.COMM_WORLD, root=0) as executor:
        if executor is not None: 
            bar = progressbar.ProgressBar(max_value=gpg.size())
            print("Computing Morse Graphs")
            results = list(bar(executor.map(work, range(0,gpg.size()),chunksize=65536)))
            SaveDatabase(outfile, results, gpg)



# ComputeSingleGeneQuery.py
# MIT LICENSE 2017
# Shaun Harker

from DSGRN.Query.Graph import *
from DSGRN.Query.Hexcodes import *
from DSGRN.Query.Logging import LogToSTDOUT

import functools

class ComputeSingleGeneQuery:
  def __init__(self, network, gene, labeller):
    self.network = network
    self.gene = gene
    self.parametergraph = ParameterGraph(self.network)
    # D is the number of network nodes
    self.D = self.parametergraph.dimension()
    self.names = [ self.network.name(i) for i in range(0, self.D)]
    # DSGRN uses an indexing scheme to refer to parameters. It is based on a mixed-radix number scheme
    # where the place value of each digit varies according to the number of logic parameters for each node
    # and the number of order parameter for each node. Specifically, the ordering of the digits is (from least
    # significant) the sizes of each factor graph, followed by the number of permutations of the out-edges for
    # each node. We call these "bases" (as in number base) and we compute the place value for each digit.
    self.indexing_place_bases = [self.parametergraph.logicsize(i) for i in range(0,self.D)] + [self.parametergraph.ordersize(i) for i in range(0,self.D)]
    self.indexing_place_values = functools.reduce ( lambda x, y : x + [x[-1]*y], self.indexing_place_bases[:-1], [1])

    # "gene_index" gives the integer index used in the representation
    self.gene_index = self.network.index(self.gene)
    # num_gene_param is the size of the factor graph associated with "gene"
    self.num_gene_param = self.indexing_place_bases[self.gene_index]
    # num_reduced_param is the product of the sizes of all remaining factor graphs, and reorderings of all genes (including "gene")
    self.num_reduced_param = self.parametergraph.size() / self.num_gene_param

    # Create factor graph
    # TODO: The following algorithm for creating a factor graph takes O(NM^2) time, where N is length of hexcodes and M is size of factorgraph.
    #       An O(NM) algorithm is possible (based on looping through codewords and trying out each promotion, using a hash table)
    #       Implement this.
    self.hexcodes = self.parametergraph.factorgraph(self.gene_index)
    self.vertices = set(range(0,len(self.hexcodes)))
    self.edges = [ (gpi1, gpi2) for gpi1 in self.vertices for gpi2 in self.vertices if isAdjacentHexcode(self.hexcodes[gpi1], self.hexcodes[gpi2]) ]
    self.graph = Graph(self.vertices,self.edges)
    self.full_parameter_index = lambda rpi, gpi, gene_index : rpi % self.indexing_place_values[gene_index] + gpi * self.indexing_place_values[gene_index] + (rpi // self.indexing_place_values[gene_index]) * self.indexing_place_values[gene_index+1]
    self.reduced_parameter_index = lambda pi, gene_index : (pi % self.indexing_place_values[gene_index] + (pi // self.indexing_place_values[gene_index+1]) * self.indexing_place_values[gene_index], (pi // self.indexing_place_values[gene_index]) % self.indexing_place_bases[gene_index] )
    self.labeller = labeller

  def __call__ (self, reduced_parameter_index):
    """
    The query returns a graph which contains the poset of gene parameter indices 
    corresponding to adjusting the parameter by changing the logic parameter associated 
    with the gene being queried. The vertices are labelled according to the function
    labeller which accepts full parameter indices.

    The graph is as follows: 

    * The vertices of the graph are named according to Gene Parameter Index (gpi). 
    * There is a directed edge p -> q iff p < q and the associated logic parameters are adjacent.
    * The graph is labelled (graph.matching_label) with the output of `labeller`

    In addition the following extra structures are provided:

    * `graph.num_inputs` is the number of network edges which are inputs to the gene associated with the query
    * `graph.num_outputs`is the number of network edges which are outputs to the gene associated with the query
    * `graph.essential` is a boolean-valued function which determines if each vertex corresponds to an essential parameter node
    """
    graph = self.graph.clone()
    graph.matching_label = lambda gpi : self.labeller(self.full_parameter_index(reduced_parameter_index,gpi,self.gene_index))
    graph.num_inputs = len(self.network.inputs(self.gene_index))
    graph.num_outputs = len(self.network.outputs(self.gene_index))
    #graph.essential = lambda gpi : essential(graph.data[gpi][0],graph.num_inputs,graph.num_outputs)
    return graph

  def number_of_gene_parameters(self):
    """
    Return number of gene parameters associated with query object
    This is the size of the logic factor graph associated with the selected gene
    """
    return self.num_gene_param

  def number_of_reduced_parameters(self):
    """
    Return number of reduced parameters associated with query object
    This is the size of the parameter grapph divided by the size of the 
    logic factor graph associated with the selected gene
    """  
    return self.num_reduced_param




# EquilibriumCells.py
# Marcio Gameiro
# MIT LICENSE
# 2021-07-03

import pychomp2

__all__ = ['EquilibriumCells']

def EquilibriumCells(parameter, eqtype='all', eqformat='coords'):
    """Return the DSGRN equilibrium cells for the given parameter.
    """
    # Check if equilibrium cell type is valid
    eqtype_vals = ['all', 'a', 'topdim', 't', 'nontop', 'n']
    if eqtype not in eqtype_vals:
        raise ValueError(f"Invalid value '{eqtype}' for eqtype. Supported values are: {', '.join(map(repr,eqtype_vals))}.")
    # Check if equilibrium output format is valid
    eqformat_vals = ['coords', 'c', 'index', 'i']
    if eqformat not in eqformat_vals:
        raise ValueError(f"Invalid value '{eqformat}' for eqformat. Supported values are: {', '.join(map(repr,eqformat_vals))}.")
    # Get network information
    network = parameter.network()
    D = network.size()
    # Use a cubical complex to access lower dimensional cells
    cc = pychomp2.CubicalComplex([n + 1 for n in network.domains()])
    labelling = parameter.labelling()
    limits = network.domains()
    pv = [1]
    for k in limits:
        pv.append(pv[-1] * k)

    def closedrightfringe(c):
        return any(cc.rightfringe(k) for k in cc.star({c}))

    def Absorbing(kappa, d, direction):
        """Return True if the codim 1 face of top cell kappa collapsed
        in dimension d in direction "direction" is absorbing. Here
        direction = -1 means left wall and direction = 1 means right wall.
        """
        coords = cc.coordinates(kappa)
        # The "wrap layer" boxes have all walls absorbing
        if any(coords[d] == limits[d] for d in range(D)):
            return True
        # The cc kappa indexing is not compatible with labelling indexing
        # due to the extra wrapped layer, so we compute the index idx
        idx = sum(c * pv[k] for (k, c) in enumerate(coords))
        labelbit = 1 << (d + (D if direction == 1 else 0))
        return labelling[idx] & labelbit != 0

    def RookField(kappa):
        """Return the rook field of a cubical top-cell kappa as a list
        of tuples, where each tuple gives the sign of the flow at the
        lef and right walls, respectively.
        """
        sign = lambda kappa, d, direc: direc if Absorbing(kappa, d, direc) else -direc
        return [(sign(kappa, d, -1), sign(kappa, d, 1)) for d in range(D)]

    def NormalVariables(sigma):
        # Inessential directions
        shape = cc.cell_shape(sigma)
        return [d for d in range(D) if shape & (1 << d) == 0]

    def TangentVariables(sigma):
        # Essential directions
        shape = cc.cell_shape(sigma)
        return [d for d in range(D) if shape & (1 << d) != 0]

    def Shape(sigma):
        """Return shape of sigma as a tuple.
        """
        shape = cc.cell_shape(sigma)
        return [1 if (shape & (1 << d) != 0) else 0 for d in range(D)]

    def RelativePositionVector(sigma, tau):
        """Return the relative position vector of sigma with respect to tau. This is
        equal to the difference of combinatorial position of two cells, regarding
        vertices to be at (even, even, ...) integer coordinates and higher dimensional
        cells to have odd coordinates for their dimensions with extent.
        """
        # Matches the definiton in the paper if sigma is a face of tau
        return [x-y for (x, y) in zip(cc.barycenter(sigma), cc.barycenter(tau))]

    def Phi(sigma, kappa):
        """Return the rook field extension from a top-cell kappa to
        a lower dimensionl cell sigma.
        """
        rf = RookField(kappa)        # Rook field of top cell
        J_i = NormalVariables(sigma) # Inessential directions
        p = RelativePositionVector(sigma, kappa)
        # Set flow to zero if essential direction and opposite signs at walls
        phi = lambda left, right, d: (left if p[d] == -1 else right) if d in J_i else 0
        return [left if left == right else phi(left, right, d) for d, (left, right) in enumerate(rf)]

    def Opaque(sigma):
        """Return True if sigma is an opaque cubical cell.
        """
        J_i = NormalVariables(sigma) # Inessential directions
        # Compute Phi(sigma, kappa) for kappa in the top star of sigma
        phi_tstar = [Phi(sigma, kappa) for kappa in cc.topstar(sigma) if not cc.rightfringe(sigma)]
        return all(n not in J_i or set(phi) == {-1, 1} for (n, phi) in enumerate(zip(*phi_tstar)))

    def isEquilibriumCell(sigma):
        """Return True if sigma is an equilibrium cubical cell.
        """
        if closedrightfringe(sigma):
            return False
        if not Opaque(sigma):
            return False
        for kappa in cc.topstar(sigma):
            rf = RookField(kappa)
            if any(rf[d][0] == rf[d][1] for d in TangentVariables(sigma)):
                return False
        return True

    # Get equilibrium cells
    if eqtype in ['all', 'a']:
        # Get all equilibrium cells
        eq_cells = [sigma for sigma in cc if isEquilibriumCell(sigma)]
    elif eqtype in ['topdim', 't']:
        # Get only the top-dimensional equilibrium cells
        eq_cells = [sigma for sigma in cc if isEquilibriumCell(sigma) if cc.cell_dim(sigma) == D]
    elif eqtype in ['nontop', 'n']:
        # Get only the lower-dimensional equilibrium cells
        eq_cells = [sigma for sigma in cc if isEquilibriumCell(sigma) if cc.cell_dim(sigma) < D]
    # else: # Unreachable

    # Return the indices if eqformat is index
    if eqformat in ['index', 'i']:
        return eq_cells

    # Return coordinates (eqformat is coords)
    if eqtype in ['topdim', 't']:
        # Return only the coordinates of lowest vertex
        equilibrium_cells = [tuple(cc.coordinates(sigma)) for sigma in eq_cells]
    else:
        # Return coordinates of lowest vertex and shape of cell
        equilibrium_cells = [[tuple(cc.coordinates(sigma)), tuple(Shape(sigma))] for sigma in eq_cells]
    return equilibrium_cells



# HysteresisQuery.py
# MIT LICENSE 2016
# Shaun Harker

from DSGRN.Query.Graph import *
from DSGRN.Query.AlignmentGraph import *
from DSGRN.Query.MonostableFixedPointQuery import *
from DSGRN.Query.SingleFixedPointQuery import *
from DSGRN.Query.DoubleFixedPointQuery import *
from DSGRN.Query.SingleGeneQuery import *
from DSGRN.Query.Logging import LogToSTDOUT

class HysteresisQuery:
  """
  HysteresisQuery
    A HysteresisQuery post-processes the result graph of a SingleGeneQuery
    and returns true if there exists a path from the root of the graph 
    to the leaf of the graph which has the following property:
      The path matches the regular expression 
        Q(Q|q)*BB*(P|p)*P, i.e. starts at Q at root, passes through Q or q zero or more times,
        then B one or more times, and then p or P zero or more times and ends up at P at leaf
      where vertices are labeled
        Q if it matches MonostableFixedPoint(quiescent_bounds) query
        P if it matches MonostableFixedPoint(proliferative_bounds) query
        q if it matches SingleFixedPointQuery(database, quiescent_bounds) query but does not match Q
        p if it matches SingleFixedPointQuery(database, proliferative_bounds) query but does not match P
        B if it matches DoubleFixedPointQuery(database, quiescent_bounds, proliferative_bounds)
        O otherwise
  """
  def __init__(self, database, gene, quiescent_bounds, proliferative_bounds):
    """
    In order to perform hysteresis queries we must first categorize each Morse graph as either 
    'Q' monostable quiescent, 'P' monostable proliferative, 'q' quiescent, 'p' proliferative, 'B' bistable, or 'O' other
    We assume the quiescent and proliferative FP states are given by disjoint bounding rectangles
    """
    LogToSTDOUT("HysteresisQuery(" + str(database.dbname) + ", " + str(gene) + ")")
    self.database = database
    self.gene = gene
    # Create query object to check if morse graph indices have quiescent FP as only minimal morse node
    LogToSTDOUT("HysteresisQuery :: MonostableFixedPointQuery(" + str(database.dbname) + ", " + str(quiescent_bounds) + ")")
    Q = MonostableFixedPointQuery(database, quiescent_bounds)
    # Create query object to check if morse graph indices has quiescent FP
    LogToSTDOUT("HysteresisQuery :: SingleFixedPointQuery(" + str(database.dbname) + ", " + str(quiescent_bounds) + ")")
    q = SingleFixedPointQuery(database, quiescent_bounds)
    # Create query object to check if morse graph indices have proliferative FP as only minimal morse node
    LogToSTDOUT("HysteresisQuery :: MonostableFixedPointQuery(" + str(database.dbname) + ", " + str(proliferative_bounds) + ")")
    P = MonostableFixedPointQuery(database, proliferative_bounds)
    # Create query object to check if morse graph indices has proliferative FP
    LogToSTDOUT("HysteresisQuery :: SingleFixedPointQuery(" + str(database.dbname) + ", " + str(proliferative_bounds) + ")")
    p = SingleFixedPointQuery(database, proliferative_bounds)
    # Check query object to check if morse graph index has both quiescent FP and proliferative FP
    LogToSTDOUT("HysteresisQuery :: DoubleFixedPointQuery(" + str(database.dbname) + ", " + str(quiescent_bounds) + ", " + str(proliferative_bounds) + ")")
    B = DoubleFixedPointQuery(database, quiescent_bounds, proliferative_bounds)
    # Create a labelling function which accepts a morse graph index and returns Q, P, B, p, q, or O
    # Note: case fallthrough order matters here
    self.matching_label = lambda mgi : 'Q' if Q(mgi) else ( 'P' if P(mgi) else ( 'B' if B(mgi) else ( 'q' if q(mgi) else ( 'p' if p(mgi) else 'O') ) ) )
    # Create the pattern graph to represent Q -> B -> P (with self-loop on Q, B, and P)
    self.patterngraph = Graph(set([0,1,2,3,4]), [(0,0),(1,1),(0,1),(1,0),(0,2),(1,2),(2,2),(2,3),(2,4),(3,3),(3,4),(4,4),(4,3)])
    self.patterngraph.matching_label = lambda v : { 0:'Q', 1:'q', 2:'B', 3:'p', 4:'P' }[v]
    # Create matching relation (in this case we just check for equality of the matching labels)
    self.matching_relation = lambda label1, label2 : label1 == label2
    # Create SingleGeneQuery object
    LogToSTDOUT("HysteresisQuery :: SingleGeneQuery(" + str(database.dbname) + ", " + str(gene) + ")")
    self.GeneQuery = SingleGeneQuery(database, gene)
    self.memoization_cache = {}

  def __call__(self, reduced_parameter_index):
    LogToSTDOUT("HysteresisQuery(" + str(reduced_parameter_index)  + ")") # DEBUG
    searchgraph = self.GeneQuery(reduced_parameter_index)
    LogToSTDOUT("HysteresisQuery: Search Graph Constructed." ) # DEBUG
    searchgraph.matching_label = lambda v : self.matching_label(searchgraph.mgi(v))
    searchgraphstring = ''.join([ searchgraph.matching_label(v) for v in searchgraph.vertices ])
    if searchgraphstring not in self.memoization_cache:
      alignment_graph = AlignmentGraph(searchgraph, self.patterngraph, self.matching_relation)
      LogToSTDOUT("HysteresisQuery: Alignment Graph Constructed." ) # DEBUG
      LogToSTDOUT("HysteresisQuery: Alignment Graph has " + str(len(alignment_graph.vertices)) + " vertices"  ) # DEBUG
      LogToSTDOUT("HysteresisQuery: Alignment Graph has " + str(len(alignment_graph.edges)) + " edges"  ) # DEBUG
      root_vertex = (0,0)
      leaf_vertex = (len(searchgraph.vertices)-1, 4)
      is_reachable = alignment_graph.reachable(root_vertex, leaf_vertex) 
      #number_of_paths = alignment_graph.numberOfPaths(root_vertex, leaf_vertex) 
      LogToSTDOUT("HysteresisQuery: Reachability computed." ) # DEBUG
      self.memoization_cache[searchgraphstring] = is_reachable
    LogToSTDOUT("HysteresisQuery: Returning." ) # DEBUG
    return self.memoization_cache[searchgraphstring]
    
